;=======================================================
; Sprite Bouncing Note Block
; By Erik
;
; This patch gives sprites the functionality to bounce
; on note blocks.
; The sprites must run object interaction for this to
; work, as well as have some sort of gravity.
; Some sprites will act weird, namely:
; - Yoshi (when he isn't on Mario)
; - Thwomps
; - Para-Goombas/Bombs
; - Jumping Piranha Plants
; - Wiggler
; - Sprite bubbles
; - Reflecting fireballs
; - Urchins
; - Volcano Lotuses (disappointing, I know)
; - Pokeys
; - Boo Streams
; Since their behavior is funny, I let them in. Go ham.
;=======================================================

!cust_noteblock = 1             ;   If 0, every vanilla note block will bounce sprites. If 1, only one specific note block.
!cnb_number     = $0242         ;   Number of the custom note block, if the above is 1. Acts like can be whatever.
!allow_mario    = 1             ;   If 0, Mario won't be able to bounce in the custom note block.
                                ;   Note: If set to 1, the acts like setting for the custom block must be 113.
                                ;   tbh tho just set cust_noteblock to 0 it makes no sense.
!cust_tile      = 1             ;   If 1, tweak the graphics the note block bounce sprite uses. Makes no sense if !cust_noteblock is 0.
!tile           = $6B           ;   New tile for the bounce sprite.
!props          = $8F           ;   New properties for the bounce sprite.
!freeram        = $7FAB70       ;\
!freeram2       = !freeram+$0C  ; | Requires 24 bytes (44 in a SA-1 ROM)
!sa1_ram        = $4000FC       ; | Should be in bank 7F for non SA-1 ROMs. In SA-1 it doesn't matter, but preferrably in banks 40 or 41.
!sa1_ram2       = !sa1_ram+$16  ;/

        !sa1            = 0
        !addr           = $0000
        !fast           = $800000
        !sprite_slots   = $0B

        !9E             = $9E
        !AA             = $AA
        !D8_ptr         = $D8,x
        !D8             = $D8
        !E4             = $E4
        !14C8           = $14C8
        !14D4           = $14D4
        !14E0           = $14E0
        !154C           = $154C
        !1588           = $1588
        !15AC           = $15AC
        !15D0           = $15D0        
if read1($00FFD5) == $23
        sa1rom
        !sa1            = 1
        !addr           = $6000
        !fast           = $000000
        !sprite_slots   = $15

        !freeram       := !sa1_ram
        !freeram2      := !sa1_ram2
        !9E             = $3200
        !AA             = $9E
        !D8_ptr         = ($CC)
        !D8             = $3216
        !E4             = $322C
        !14C8           = $3242
        !14D4           = $3258
        !14E0           = $326E
        !154C           = $32DC
        !1588           = $334A
        !15AC           = $338C
        !15D0           = $754C
endif

if !sa1 = 0
        org $008023
                autoclean JML clear_7Fxxxx
endif
org $01940A
        autoclean JML check_note_block_up
org $019533
        autoclean JSL gps_fix_a
org $0290E1
        autoclean JML bounce_sprite_hijack
org $0293C3
        autoclean JSL interaction_fix
        BRA $00
org $0192A7
        autoclean JML interaction_fix_b
org $019098
        autoclean JML interaction_fix_c
if !cust_noteblock = 1
        if !allow_mario = 1
                org $029159
                        TSB $1402|!addr
                org $00C5B3
                        dw $C7F8
                org $00F19B
                        autoclean JML set_custom_block
                org $02916A
                        BNE +
                org $02917D
                +       autoclean JML reset_custom_block
                -       RTS
                org $00C58F
                        autoclean JML clear_custom_block
                org $029199
                        autoclean JML stop_custom_block
                org $029161
                        BNE -
                if !cust_tile = 1
                        org $028823
                                autoclean JSL set_custom_bounce_props
                        org $029251
                                autoclean JSL set_custom_bounce_tile
                endif
        endif
endif

freedata
gps_fix_a:
        PHA
        LDA $05
        STA $F6
        LDA $06
        STA $F7
        LDA $07
        STA $F8
        PLA
        JSL $06F700|!fast
        RTL

interaction_fix:
        ORA !15D0,x
        ORA !154C,x
        ORA !freeram,x
        RTL

interaction_fix_b:
        LDA $1693|!addr
        STA $18A7|!addr
        LDA !freeram,x
        BEQ +
        LDA !1588,x
        AND #$FC
        STA !1588,x
+       JML $0192AD|!fast

interaction_fix_c:
        LDA !15AC,x
        ORA !freeram,x
        BNE +
        JML $01909D|!fast
+       JML $0190B1|!fast

if !sa1 = 0
        clear_7Fxxxx:
                LDA.w #$01FF
                TCS
                STZ $2181
                SEP #$10
                LDY #$01
                STY $2183
                LDA.w #$8008
        .zero_byte
                STA $4300
                LDA.w #.zero_byte+1
                STA $4302
                LDX.b #.zero_byte>>16
                STX $4304
                STZ $4305
                STY $420B
                REP #$10
                JML $808027
endif

check_note_block_up:
        LDA !9E,x
        CMP #$45
        BEQ .return
        CMP #$2E
        BEQ .return
        CMP #$2B
        BEQ .return
        LDA [$F6]
        LDY $1693|!addr
        if !cust_noteblock = 0
                CMP #$01
                BNE .return
                CPY #$11
                BEQ .spawn_bounce
                CPY #$13
                BEQ .spawn_bounce
                CPY #$15
                BEQ .spawn_bounce
                CPY #$16
                BEQ .spawn_bounce
                CPY #$52
                BNE .return
                JML $019424|!fast
        else
                CPY #$52
                BNE +
                JML $019424|!fast
        +       CMP.b #!cnb_number>>8
                BNE .return
                XBA
                DEC $F8
                LDA [$F6]
                CMP.b #!cnb_number
                BNE .return
        endif
.spawn_bounce
        LDA !freeram,x
        INC
        STA !freeram,x
        LDX #$0E
        LDA #$02
        LDY #$03
        JSR spawn_bounce_sprite
        LDX $15E9|!addr
        LDA !freeram2,x
        BEQ +
        ASL #2
        STA $00
        TYA
        ORA $00
        STA !freeram2,x
        JML $019424|!fast
+       TYA
        STA !freeram2,x
        JML $019424|!fast
.return
        LDA !D8_ptr
        AND #$F0
        JML $01940E|!fast

bounce_sprite_hijack:
        LDA $1699|!addr,x
        EOR #$02
        ORA $9D
        BEQ +
        JML $0290CD|!fast
+       TXY
        LDX #!sprite_slots      ;   load slots
--      LDA !freeram,x          ;\  0 - not a sprite in it
        STA $01                 ;/
        BEQ .check_next         ;
-       XBA                     ;   preserve amount of note blocks the sprite is in
        LDA !freeram2,x         ;\
        AND #$03                ; | only get the first two bits in scratch
        STA $00                 ;/
        XBA                     ;   restore
        CMP #$01                ;\  branch if only on one note block
        BEQ +                   ;/

        LDA !freeram2,x         ;\
        LSR #2                  ; | get bits 3 and 4 and shift them right twice
        STA $00                 ;/
        DEC $01                 ;   reduce extra note blocks checked
        TYA
        CMP $00
        BEQ found_sprite
        LDA $01
        BRA -

+       TYA                     ;\  compare index
        CMP $00                 ;/
        BEQ found_sprite        ;   branch if correct
.check_next
        DEX
        BPL --
        LDX $1698|!addr
.return
        JML $0290E5|!fast

found_sprite:
        STX $15E9|!addr
        LDX $1698|!addr
        if !cust_tile   = 1
                LDY $91ED,x
                LDA.b #!tile
                STA $0202|!addr,y
        endif
;   based on original code around $0290E5
        LDA $169D|!addr,x
        BNE +
        INC $169D|!addr,x
        PHK
        PEA.w .jsl_return_1-1
        PEA $B888
        PEA $91B7
        JML $029265|!fast
.jsl_return_1
        LDA $16C9|!addr,x
        AND #$03
        TAY
        LDA $90D6|!addr,y
        CMP #$80
        BEQ .no_y
        LDX $15E9|!addr
        STA !AA,x
        LDX $1698|!addr
.no_y
        LDA $90DA|!addr,y
        CMP #$80
        BEQ +
        LDX $15E9|!addr
        STA $B6,x
        LDX $1698|!addr
+       PHK
        PEA.w .jsl_return_2-1
        PEA $B888
        PEA $B525
        JML $02B51A|!fast
.jsl_return_2
        LDA $16C9|!addr,x
        AND #$03
        TAY
        LDA $16B1|!addr,x
        CLC
        ADC $90CE,y
        STA $16B1|!addr,x
        LDA $16B5|!addr,x
        CLC
        ADC $90D2,y
        STA $16B5|!addr,x
        TXY
        LDX $15E9|!addr
        LDA $16A1|!addr,y
        SEC
        SBC #$10
        STA !D8,x
        LDA $16A9|!addr,y
        SBC #$00
        STA !14D4,x
        STZ !AA,x
        LDA $16C5|!addr,y
        BNE return_handler
        LDA #$A0
        STA !AA,x
        LDA !D8,x
        SEC
        SBC #$02
        STA !D8,x
        LDA !14D4,x
        SBC #$00
        STA !14D4,x
        LDA #$08
        STA $1DFC|!addr
        LDA !freeram,x
        DEC
        STA !freeram,x
        BEQ +
        LDA !freeram2,x
        AND #$03
+       STA !freeram2,x
        if !cust_noteblock = 0
                TYX
                PHK
                PEA.w .jsl_return_3-1
                PEA $B888
                JML $02919F|!fast
        .jsl_return_3
        else
        reset_custom_tile:
                LDA $16A5|!addr,y
                CLC
                ADC #$08
                AND #$F0
                STA $9A
                LDA $16AD|!addr,y
                ADC #$00
                STA $9B
                LDA $16A1|!addr,y
                CLC
                ADC #$08
                AND #$F0
                STA $98
                LDA $16A9|!addr,y
                ADC #$00
                STA $99
                REP #$10
                LDX #!cnb_number
                JSR change_map16
                SEP #$10
                TYX
        endif
        JML $0290CA|!fast
return_handler:
        LDX $1698|!addr
        JML $0290CD|!fast

if !cust_noteblock = 1
        if !allow_mario = 1
                reset_custom_block:
                        LDA $16C9|!addr,x
                        AND #$03
                        CMP #$03
                        BNE +
                        LDA #$08
                        STA $1DFC|!addr
                +       LDA $1402|!addr
                        BMI +
                        JML $029182|!addr
                +       STZ $1402|!addr
                        TXY
                        BRA reset_custom_tile
                set_custom_block:
        lda $1340
                        LDA $01,s
                        PHX
                        TAX
                        LDA !14C8,x
                        CMP #$09
                        BCC .check_mario
                        CMP #$0A
                        BNE +
                        LDA $99
                        STA $09
                        LDA $9B
                        STA $0B
                        LDA $98
                        AND #$F0
                        STA $08
                        LDA $9A
                        LSR #4
                        ORA $08
                        BRA .start_check
                +       LDA !14D4,x
                        STA $09
                        LDA !14E0,x
                        STA $0B
                        LDA !D8,x
                        AND #$F0
                        STA $08
                        LDA !E4,x
                        BIT #$08
                        BEQ +
                        CLC
                        ADC #$10
                +       LSR #4
                        ORA $08
                .start_check
                        XBA
                        LDA $5B
                        LSR
                        XBA
                        BCC .horz
                        LDX $09
                        CLC
                        ADC $00BA80,x
                        STA $0C
                        LDA $00BABC,x
                        ADC $0B
                        STA $0D
                        BRA .common

                .horz
                        LDX $0B
                        ADC $00BA60,x
                        STA $0C
                        LDA $00BA9C,x
                        ADC $09
                        STA $0D

                .common
                        if !sa1 = 1
                                LDA #$41
                        else
                                LDA #$7F
                        endif
                        STA $0E
                        LDA [$0C]
                        CMP.b #!cnb_number>>8
                        BNE .check_mario
                        DEC $0E
                        LDA [$0C]
                        CMP.b #!cnb_number
                        BEQ .sprite
                .check_mario
                        LDA [$00]
                        CMP.b #!cnb_number>>8
                        BNE .not_set
                        DEC $02
                        LDA [$00]
                        CMP.b #!cnb_number
                        BNE .not_set
                .sprite
                        LDA #$80
                        TSB $1402|!addr
                .not_set
                        PLX
                        LDA $00F080,x
                        JML $00F19F|!fast
                clear_custom_block:
                        LDA #$01
                        TRB $1402|!addr
                        JML $00C7F8|!fast
                stop_custom_block:
                        STZ $1699|!addr,x
                        STZ $1402|!addr
                        JML $0290CD|!fast
                if !cust_tile = 1
                        set_custom_bounce_props:
                                TAX
                                PHY
                                LDA $8789,x
                                LDY $1402|!addr
                                BPL +
                                LDA.b #!props
                        +       PLY
                                RTL
                        set_custom_bounce_tile:
                                TAX
                                PHY
                                LDA $91F0,x
                                LDY $1402|!addr
                                BPL +
                                LDA.b #!tile
                        +       PLY
                                RTL
                endif
        endif
endif

;=================================
  spawn_bounce_sprite:
; Based on GPS' routine.
;
; Input:
;   A = Bounce sprite number
;   X = $9C value
;   Y = bounce sprite direction
; Ouput:
;   Y = Slot used
;=================================

        STA $04
        STY $06
        STX $07
        LDA #$0F
        TRB $0A
        TRB $0C
        LDY #$03
.find_free        
        LDA $1699|!addr,y
        BEQ .slot_found
        DEY 
        BPL .find_free
        DEC $18CD|!addr
        BPL .alt_slot
        LDA #$03
        STA $18CD|!addr
.alt_slot        
        LDY $18CD|!addr
        LDA $1699|!addr,y
        CMP #$07
        BNE .no_turn_block_reset
        LDA $04 : PHA 
        PEI ($06)
        PEI ($0A)
        PEI ($0C)
        LDA $16A5|!addr,y
        STA $0A
        LDA $16AD|!addr,y
        STA $0B
        LDA $16A1|!addr,y
        CLC 
        ADC #$0C
        AND #$F0
        STA $0C
        LDA $16A9|!addr,y
        ADC #$00
        STA $0D
        LDA $16C1|!addr,y
        STA $9C
        JSL $00BEB0|!fast
        REP #$20
        PLA 
        STA $0C
        PLA 
        STA $0A
        PLA 
        STA $06
        SEP #$20
        PLA 
        STA $04
.no_turn_block_reset 
        LDY $18CD|!addr
.slot_found        
        LDA $04
        STA $1699|!addr,y
        LDA #$00
        STA $169D|!addr,y
        LDA $0A
        STA $16A5|!addr,y
        LDA $0B
        STA $16AD|!addr,y
        LDA $0C
        STA $16A1|!addr,y
        LDA $0D
        STA $16A9|!addr,y
        LDA $1933|!addr
        LSR
        ROR
        STA $08
        LDX $06
        LDA.l $02873A,x
        STA $16B1|!addr,y
        LDA.l $02873E,x
        STA $16B5|!addr,y
        TXA
        ORA $08
        STA $16C9|!addr,y
        LDA $07
        STA $16C1|!addr,y
        LDA #$08
        STA $16C5|!addr,y
        if !cust_tile = 1
                LDA.b #!props
        else
                LDA.b #read1($02878A)
        endif
        STA $1901|!addr,y

        PHY
        LDY #$03
-       LDA $16CD|!addr,y
        BEQ +
        DEY
        BPL -
        INY
+       LDA $0A
        STA $16D1|!addr,y
        LDA $0B
        STA $16D5|!addr,y
        LDA $0C
        STA $16D9|!addr,y
        LDA $0D
        STA $16DD|!addr,y
        LDA $1933|!addr
        BEQ .layer_1
        LDA $0A
        SEC
        SBC $26
        STA $16D1|!addr,y
        LDA $0B
        SBC $27
        STA $16D5|!addr,y
        LDA $0C
        SEC
        SBC $28
        STA $16D9|!addr,y
        LDA $0D
        SBC $29
        STA $16DD|!addr,y
.layer_1
        LDA #$01
        STA $16CD|!addr,y
        LDA #$06
        STA $18F8|!addr,y
        PLY
        RTS

if !cust_noteblock = 1
        ;====================================
          change_map16: ; Optimized version |
        ; By Akaginite                      |
        ;====================================

                PHP
                PHB
                PHY
                PHX
                LDY $98
                STY $0E
                LDY $9A
                STY $0C
                SEP #$30
                LDA $5B
                LDX $1933|!addr
                BEQ .Layer1
                LSR
        .Layer1
                STA $0A
                LSR
                BCC .Horz
                LDA $9B
                LDY $99
                STY $9B
                STA $99
        .Horz
                LDA $99
                CMP #$02
                BCC .NoEnd
                REP #$10
                PLX
                PLY
                PLB
                PLP
                RTS

        .NoEnd
                LDA $9B
                STA $0B
                ASL
                ADC $0B
                TAY
                REP #$20
                LDA $98
                AND #$01F0
                STA $08
                AND #$00F0
                ASL #2          ; 0000 00YY YY00 0000
                XBA             ; YY00 0000 0000 00YY
                STA $06
                TXA
                SEP #$20
                ASL
                TAX
                
                LDA $0D
                LSR
                LDA $0F
                AND #$01        ; 0000 000y
                ROL             ; 0000 00yx
                ASL #2          ; 0000 yx00
                ORA #$20        ; 0010 yx00
                CPX #$00
                BEQ .NoAdd
                ORA #$10        ; 001l yx00
        .NoAdd
                TSB $06         ; $06 : 001l yxYY
                LDA $9A         ; X LowByte
                AND #$F0        ; XXXX 0000
                LSR #3          ; 000X XXX0
                TSB $07         ; $07 : YY0X XXX0
                LSR
                TSB $08

                LDA $1925|!addr
                ASL
                REP #$31
                ADC $00BEA8,x
                TAX
                TYA
                if !sa1
                        ADC $000000,x
                        TAX
                        LDA $08
                        ADC $000000,x
                else
                        ADC $00,x
                        TAX
                        LDA $08
                        ADC $00,x
                endif
                TAX
                PLA
                ASL
                TAY
                LSR
                SEP #$20
                if !sa1
                        STA $400000,x
                        XBA
                        STA $410000,x
                else
                        STA $7E0000,x
                        XBA
                        STA $7F0000,x
                endif
                LSR $0A
                LDA $1933|!addr
                REP #$20
                BCS .Vert
                BNE .HorzL2
        .HorzL1
                LDA $1A         ;\
                SBC #$007F      ; |$08 : Layer1X - 0x80
                STA $08         ;/
                LDA $1C         ;  $0A : Layer1Y
                BRA +
        .HorzL2
                LDA $1E         ;\ $08 : Layer2X
                STA $08         ;/
                LDA $20         ;\ $0A : Layer2Y - 0x80
                SBC #$007F      ;/
                BRA +

        .Vert
                BNE .VertL2
                LDA $1A         ;\ $08 : Layer1X
                STA $08         ;/
                LDA $1C         ;\ $0A : Layer1Y - 0x80
                SBC #$0080      ;/
                BRA +
        .VertL2
                LDA $1E         ;\
                SBC #$0080      ; |$08 : Layer2X - 0x80
                STA $08         ;/
                LDA $20         ;  $0A : Layer2Y
        +       STA $0A
                PHK
                PEA.w .return-1
                PEA $804C
                JML $00C0FB|!fast
        .return
                PLY
                PLB
                PLP
                RTS
endif

